/*
 * Copyright (C) 2012 Michael Andersen <michael@steelcode.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.steelcode.pamflit;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.NoSuchElementException;
import java.util.Scanner;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * A crude hack for a PVM parser.
 * This file is particularly ugly and filled with bad parts. Please don't judge me.
 * 
 * @author Michael Andersen <michael@steelcode.com>
 */
public class PVMParser
{
    private Scanner in;
    private ArrayList<String> stringPool;
    private HashMap<String, ArrayList<Integer>> stringRefs;
    private HashMap<String, Integer> labelDecs; //For when addresses are found
    private HashMap<String, ArrayList<Integer>> labelRefs; //For unresovled labels
    private ArrayList<Integer> codes;
    //zomg what a long regex.
    private static Pattern standardlineRegex = Pattern.compile(
            "^(?:([a-zA-Z_][\\w]*)\\s*:)?\\s*(\\{\\s*[0-9]*\\s*\\})?\\s*([a-zA-Z]+)?\\s*(\\-?\\w+|\".*?\")?\\s*(\\w+|\".*?\")?\\s*(?:;(.*))?$");
    private PVMParser(Scanner s)
    {
        in = s;
        codes = new ArrayList<Integer>();
        labelRefs = new HashMap<String, ArrayList<Integer>>();
        stringRefs = new HashMap<String, ArrayList<Integer>>();
        stringPool = new ArrayList<String>();
        labelDecs = new HashMap<String, Integer>();
    }
    private void emit(String mnemonic)
    {
        if(!lookup.containsKey(mnemonic.toLowerCase()))
        {
            System.err.println("Unknown mnemonic: " + mnemonic);
            System.exit(1);
        }
        codes.add(lookup.get(mnemonic.toLowerCase()));
        //System.err.println(lookup.get(mnemonic.toLowerCase()));
    }
    private int pc()
    {
        return codes.size();
    }
    private void patchStrings()
    {
        for(String s : stringPool)
        {
            int org = codes.size();
            for (int i = 0; i < s.length(); i++)
            {
                codes.add((int)s.charAt(i));
            }
            codes.add(0);
            for (Integer i : stringRefs.get(s))
            {
                codes.set(i, org);
            }
        }
    }
    private void patchLabels()
    {
        for(String s : labelRefs.keySet())
        {
            if (!labelDecs.containsKey(s))
            {
                System.err.println("Unresolved symbol: "+s);
                System.exit(1);
            }
            for (Integer i : labelRefs.get(s))
            {
                codes.set(i, labelDecs.get(s));
            }
        }
    }
    /**
     * For those reading this who have done the compiler course: I know I can do better than a regex
     * but a regex is simpler and faster (to code) for the mean time.
     * @param s The line
     */
    private void parseLine(String s)
    {
        String label        = null,
               preComment   = null,
               mnemonic     = null,
               op1          = null,
               op2          = null,
               postComment  = null;
        
        Matcher m = standardlineRegex.matcher(s);
        if (m.matches())
        {
            label = m.group(1);
            preComment = m.group(2);
            mnemonic = m.group(3);
            op1 = m.group(4);
            op2 = m.group(5);
            postComment = m.group(6);
            //System.out.println(label + " / "+ preComment + " / " + mnemonic + " / " + op1 + " / " + op2 + " / " + postComment);
            
            if(label != null) //TODO
            {
                if (labelDecs.containsKey(label))
                {
                    throw new RuntimeException("Duplicate label: "+label);
                }
                labelDecs.put(label, pc());
            }
            if(mnemonic == null)
            {
                //A comment line or standalone label of some kind
                return;
            }
            emit(mnemonic);
            if(op1 == null)
                return;
            //Strings
            if (op1.startsWith("\""))
            {
                String v = op1.substring(1, op1.length()-1);
                //v = v.toLowerCase();
                //TODO: do this properly so that "\\n is recognised as NOT \n"
                v = v.replaceAll("\\\\n", "\n");
                stringPool.add(v);
                if (!stringRefs.containsKey(v))
                {
                    stringRefs.put(v, new ArrayList<Integer>());
                }
                stringRefs.get(v).add(pc());
                codes.add(0); //placeholder
            }
            //Integers
            else if (op1.toLowerCase().startsWith("0x"))
            {
                codes.add(Integer.parseInt(op1.substring(2), 16));
            }
            else if (op1.toLowerCase().startsWith("0b"))
            {
                codes.add(Integer.parseInt(op1.substring(2),2));
            }
            else if (Character.isLetter(op1.toLowerCase().charAt(0)))
            { 
               //String l = op1.toLowerCase();
               if (labelDecs.containsKey(op1))
               {
                   codes.add(labelDecs.get(op1));
               }
               else
               {
                   if(!labelRefs.containsKey(op1))
                   {
                       labelRefs.put(op1,new ArrayList<Integer>());
                   }
                   labelRefs.get(op1).add(pc());
                   codes.add(0); //Placeholder
               }
            }
            else
            { //Integer
                codes.add(Integer.parseInt(op1));
            }
        }
        else
        {
            //TODO: This complains about perfectly valid comment-only lines or blank lines
            System.out.println("This crude parser didn't match the line '"+s+"'");
        }
    }
    private void parseFile()
    {
        while(true)
        {
            try
            {
                String l = in.nextLine().trim().toUpperCase();
                if(l.length() == 0)
                    continue;
                if (l.equals("ASSEM"))
                {
                    break;
                }
            }
            catch(NoSuchElementException e)
            {
                e.printStackTrace();
                System.err.println("'ASSEM' expected");
                return;
            }
        }
        while(true)
        {
            try
            {
                String l = in.nextLine().trim().toUpperCase();
                if(l.length() == 0)
                    continue;
                if (l.equals("BEGIN"))
                {
                    break;
                }
            }
            catch(NoSuchElementException e)
            {
                System.err.println("'BEGIN' expected");
                return;
            }
        }
        //Code:
        while(true)
        {
            try
            {
                String l = in.nextLine().trim();
                if (l.toUpperCase().startsWith("END."))
                {
                    return;
                }
                parseLine(l);
            }
            catch(Exception e)
            {
                e.printStackTrace();
                System.err.println("END. expected");
            }
        }
    }
    static public ArrayList<Integer> parse(String filename)
    {
        //A line looks like this:
        //label: { pre-comment } mnemonic operand operand ; comment
        // or like this
        //.preprocessor operands ; comment
        //The label can be ommited, as can the pre-comment
        //One or both operands can be omitted
        //operands can be decimal, hex (0x) or binary (0b)
        
        File f = new File(filename);
        Scanner in = null;
        try
        {
            in = new Scanner(f);
        }
        catch(FileNotFoundException e)
        {
            System.out.println("ERR: can't find the file to program!");
            System.exit(1);
        }
        PVMParser p = new PVMParser(in);
        p.parseFile();
        p.patchStrings();
        p.patchLabels();
        return p.codes;      
    }
    
    /*
     * These are the opcodes implemented on firmware version 1.0
     * Not all of them have been tested, so you have been warned.
     * The edge cases especially need some stress-testing (overflows etc)
     */
    static final HashMap<String, Integer> lookup; 
    static
    {
        lookup = new HashMap<String,Integer>();
        lookup.put("nop",   1);
        lookup.put("dsp",   2);
        lookup.put("ldc",   3);
        lookup.put("lda",   4);
        lookup.put("ldv",   5);
        lookup.put("sto",   6);
        lookup.put("ldxa",  7);
        lookup.put("inpi",  8);
        lookup.put("prni",  9);
        lookup.put("inpb",  10);
        lookup.put("prnb",  11);
        lookup.put("prns",  12);
        lookup.put("prnl",  13);
        lookup.put("neg",   14);
        lookup.put("add",   15);
        lookup.put("sub",   16);
        lookup.put("mul",   17);
        lookup.put("div",   18);
        lookup.put("rem",   19);
        lookup.put("not",   20);
        lookup.put("and",   21);
        lookup.put("or",    22);
        lookup.put("ceq",   23);
        lookup.put("cne",   24);
        lookup.put("clt",   25);
        lookup.put("cle",   26);
        lookup.put("cgt",   27);
        lookup.put("cge",   28);
        lookup.put("brn",   29);
        lookup.put("bze",   30);
        lookup.put("anew",  31);
        lookup.put("halt",  32);
        lookup.put("stk",   33);
        lookup.put("ldga",  34);
        lookup.put("fhdr",  35);
        lookup.put("call",  36);
        lookup.put("retv",  37);
        lookup.put("ret",   38);
        lookup.put("trap",  39);
        lookup.put("alen",  40);
        lookup.put("bfalse",41);
        lookup.put("btrue", 42);
        lookup.put("ldl",   43);
        lookup.put("stl",   44);
        lookup.put("ldg",   45);
        lookup.put("stg",   46);
        lookup.put("inc",   47);
        lookup.put("dec",   48);
        lookup.put("indx",  49);
        lookup.put("ass",   50);
        lookup.put("stoc",  51);
        lookup.put("inpc",  52);
        lookup.put("prnc",  53);
        lookup.put("incc",  54);
        lookup.put("decc",  55);
        lookup.put("stlc",  56);
        lookup.put("stgc",  57);
        lookup.put("dup",   58);
        lookup.put("lde",   59);
        lookup.put("ste",   60);
        lookup.put("stec",  61);
        lookup.put("swi",   62);
        lookup.put("sel",   63);
        lookup.put("pop",   64);
        lookup.put("iret",  65);

        //Added in firmware version 2.2
        lookup.put("lda_0",     66);                                                                         
        lookup.put("lda_1",     67);                                                                         
        lookup.put("lda_2",     68);                                                                         
        lookup.put("lda_3",     69);                                                                         
        lookup.put("lda_4",     70);                                                                         
        lookup.put("lda_5",     71);                                                                         
        lookup.put("ldga_0",    72);                                                                         
        lookup.put("ldga_1",    73);                                                                         
        lookup.put("ldga_2",    74);                                                                         
        lookup.put("ldga_3",    75);                                                                         
        lookup.put("ldga_4",    76);                                                                         
        lookup.put("ldga_5",    77);                                                                         
        lookup.put("ldl_0",     78);                                                                         
        lookup.put("ldl_1",     79);                                                                         
        lookup.put("ldl_2",     80);                                                                         
        lookup.put("ldl_3",     81);                                                                         
        lookup.put("ldl_4",     82);                                                                         
        lookup.put("ldl_5",     83);                                                                         
        lookup.put("ldg_0",     84);                                                                         
        lookup.put("ldg_1",     85);                                                                         
        lookup.put("ldg_2",     86);                                                                         
        lookup.put("ldg_3",     87);                                                                         
        lookup.put("ldg_4",     88);                                                                         
        lookup.put("ldg_5",     89);                                                                         
        lookup.put("stl_0",     90);                                                                         
        lookup.put("stl_1",     91);                                                                         
        lookup.put("stl_2",     92);                                                                         
        lookup.put("stl_3",     93);                                                                         
        lookup.put("stl_4",     94);                                                                         
        lookup.put("stl_5",     95);                                                                         
        lookup.put("stg_0",     96);                                                                         
        lookup.put("stg_1",     97);                                                                         
        lookup.put("stg_2",     98);                                                                         
        lookup.put("stg_3",     99);                                                                         
        lookup.put("stg_4",    100);                                                                         
        lookup.put("stg_5",    101);                                                                         
        lookup.put("ldc_m1",   102);                                                                         
        lookup.put("ldc_0",    103);                                                                         
        lookup.put("ldc_1",    104);                                                                         
        lookup.put("ldc_2",    105);                                                                         
        lookup.put("ldc_3",    106);                                                                         
        lookup.put("ldc_4",    107);                                                                         
        lookup.put("ldc_5",    108);                                                                         
        lookup.put("prnfi",    109);                                                                         
        lookup.put("prnfb",    110);                                                                         
        lookup.put("prnfc",    111);                                                                         
        lookup.put("max",      112);                                                                         
        lookup.put("max2",     113);                                                                         
        lookup.put("min",      114);                                                                         
        lookup.put("min2",     115);                                                                         
        lookup.put("i2c",      116);                                                                         
        lookup.put("low",      117);                                                                         
        lookup.put("cap",      118);                                                                         
        lookup.put("islet",    119);                                                                         
        lookup.put("isdig",    120);                                                                         
        lookup.put("isldg",    121);                                                                         
        lookup.put("abs",      122);                                                                         
        lookup.put("sqr",      123);                                                                         
        lookup.put("isp",      124);                                                                         
        lookup.put("sfu",      125);                                                                         
        lookup.put("sfd",      126);                                                                         
        lookup.put("efu",      127);                                                                         
        lookup.put("efd",      128);                                                                         
        lookup.put("ston",     129);                                                                         
        lookup.put("memb",     130);                                                                         
        lookup.put("sfl",      131);                                                                         
        lookup.put("tfl",      132);                                                                         
        lookup.put("efl",      133);  
        //Not added yet, but mnemonics reserved
        //lookup.put("inpstr",   134);                                                                         
        //lookup.put("prnstr",   135);                                                                         
        //lookup.put("ceqs",     136);                                                                         
        //lookup.put("cnes",     137);                                                                         
        //lookup.put("clts",     138);                                                                         
        //lookup.put("cles",     139);                                                                         
        //lookup.put("cgts",     140);                                                                         
        //lookup.put("cges",     141);                                                                         
        //lookup.put("leng",     142);                                                                         
        //lookup.put("concat",   143);                                                                         
        //lookup.put("substr",   144);                                                                         
        //lookup.put("lower",    145);                                                                         
        //lookup.put("upper",    146);                                                                         
        //lookup.put("prnfstr",  147);                                                                         
        lookup.put("ldr",      148);
        lookup.put("str",      149);
        lookup.put("inpl",     150);
        lookup.put("stof",     151);
        lookup.put("nul",      255);
    }
}
